@@ -1,6 +1,6 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2013,2014,2015,2016,2017,2018,2019,2021
-;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
+;;  Copyright (C) 2013-2025
+;;      "Mu Lei" known as "NalaGinrut" <mulei@gnu.org>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
 ;;  Lesser General Public License published by the Free Software
@@ -119,6 +119,9 @@ db.pool = increase | fixed")
     ((db encodeparams)
      #f
      "Whether to encode params automatically.
+(params rc \"your_key\") will be encoded by uri-encode.
+The username and passwd in :auth will be encoded by uri-encode.
+NOTE: It's your duty to call uri-decode to get proper value.
 NOTE: If you enable db.encodeparams then it's better to decode the related value
 twice in the client-side, since some requests may be sent from browsers, and
 they're already encoded.
@@ -226,8 +229,20 @@ port. When requests come, the Linux kernel will do the necessary lock and alloca
 work for us to dispatch requests to these Artanis instances.
 server.multi = <boolean>")
 
+    ((server workers)
+     1
+     "To specify the number of workers in the server core. Each worker implies a coroutine queue.
+The default value is 1, which means no threading at all.
+The throughput will be dramatically increased when you set it to a larger number.
+[Experimental] It's not well tested yet, but it does work. Here're known issues:
+1. The main thread may be halted when ctrl+c is pressed, but it still work with other threads.
+2. There could be potential race conditions, we've already use parameters for its functional nature
+   to implement thread local storage. But need more tests.
+NOTE: If the workers is larger than 1, then it implies server.multi = true.
+server.workers = <integer>")
+
     ((server websocket)
-     #t
+     #f
      "Enable WebSocket.
 server.websocket = <boolean>")
 
@@ -308,12 +323,23 @@ host.port = <integer>")
      "Specify the protocol family.
 host.family = ipv4 | ipv6")
 
+    ((host detectpath)
+     #f
+     "Detect host or domain-name from specified path. This is useful when you run GNU Artanis inside a container.
+host.detectpath = <PATH>|<boolean>")
+
     ;; for session namespace
     ((session path)
      "session"
      "Specify the session files path. Change according to your session engine.
 session.path = <PATH>")
 
+    ((session hijackcheck)
+     "false"
+     "Check the valid session, if it appears in multiple client IP, then fail. This may
+cause problems when your hosting is behind CDN, you may need to disable it if so.
+session.hijackcheck = <boolean>")
+
     ((session backend)
      simple
      "Specify session engine. Here're supported backends:
@@ -323,54 +349,61 @@ file: stores session information into text files.
 redis: uses Redis for managing sessions.
 session.engine = simple | db | file | redis | redis@addr:port | <third-party-engine>")
 
+    ((session i18n)
+     json
+     "Specify the i18n engine. Here're supported backends:
+1. json: uses json file for i18n.
+2. po: uses PO with gettext for i18n.
+session.i18n = json | po | <third-party-engine>")
+
     ((upload types)
      (jpg png gif)
      "Specify allowed upload file type, say, upload.types = jpg,png,gif.
-upload.types = <item-list>")
+     upload.types = <item-list>")
 
     ((upload path)
      "upload"
      "The path to put the uploaded files.
-upload.path = <PATH>")
+     upload.path = <PATH>")
 
     ((upload size)
      5242880  ; 5MB
      "The size limitation of uploaded file in bytes.
-upload.size = <interger>")
+     upload.size = <interger>")
 
     ;; for mail namespace
     ((mail sender) "/usr/bin/msmtp"
      "The command called by Sendmail module. It's strongly recommended to use `msmtp' to replace `sendmail'.
-mail.sender = <string>")
+     mail.sender = <string>")
 
     ((cache maxage)
      3600
      "The maximum age of a cached page in seconds.
-This is the global maxage of any cache.
-If you want to specify maxage for certain page, please read the manual about the Cache.
-cache.maxage = <integer>")
+     This is the global maxage of any cache.
+     If you want to specify maxage for certain page, please read the manual about the Cache.
+     cache.maxage = <integer>")
 
     ((debug enable)
      #f
      "Wheather to enable debug mode.
-If you enable debug mode, Artanis will print debug information verbosely.
-The module you modified will be reloaded instantly, and the page view will be rendered as well.
-NOTE: This option will affect the performance, please use it for debug purposes only./
-debug.enable = <boolean>")
+     If you enable debug mode, Artanis will print debug information verbosely.
+     The module you modified will be reloaded instantly, and the page view will be rendered as well.
+     NOTE: This option will affect the performance, please use it for debug purposes only./
+     debug.enable = <boolean>")
 
     ((debug monitor)
      ()
      "The paths that needs to be monitored in debug-mode.
-This will take advantage of `inotify' in GNU/Linux kernel.
-NOTE: We may support GNU/Hurd as well, with its file monitor mechanism, in the future.
-debug.monitor = <PATHs>")
- ((cookie expires)
-       3600
-      "Cookie expiration time in seconds.
-       1 hour is 3600
-       6 hours 21600
-       1 month 2592000
- cookie.expires = <integer>")))
+     This will take advantage of `inotify' in GNU/Linux kernel.
+     NOTE: We may support GNU/Hurd as well, with its file monitor mechanism, in the future.
+     debug.monitor = <PATHs>")
+    ((cookie expires)
+     3600
+     "Cookie expiration time in seconds.
+     1 hour is 3600
+     6 hours 21600
+     1 month 2592000
+     cookie.expires = <integer>")))
 
 ;; Init all fields with default values
 (for-each (lambda (x) (conf-set! (car x) (cadr x))) (default-conf-values))
@@ -424,6 +457,12 @@ debug.monitor = <PATHs>")
       #f
       x))
 
+(define-syntax-rule (->string/boolean x)
+  (case (string->symbol (string-downcase x))
+    ((true on yes enable) #t)
+    ((false off no disable) #f)
+    (else x)))
+
 (define-syntax-rule (->dbd x)
   (let ((d (->symbol x)))
     (if (eq? 'mariadb d)
@@ -444,6 +483,12 @@ debug.monitor = <PATHs>")
                 #f
                 "Invalid db.pool value, we accept: ~{~a~^,~}" *pool-modes*)))))
 
+(define-syntax-rule (->workers x)
+  (let ((w (string->number x)))
+    (cond
+     ((positive? w) w)
+     (else (error "Invalid server.workers number! Must be >= 1" x)))))
+
 (define (parse-namespace-db item)
   (match item
     (('enable usedb) (conf-set! '(db enable) (->bool usedb)))
@@ -475,6 +520,7 @@ debug.monitor = <PATHs>")
     (('bufsize bufsize) (conf-set! '(server bufsize) (->integer bufsize)))
     (('impl impl) (conf-set! '(server impl) (string->symbol impl)))
     (('multi multi) (conf-set! '(server multi) (->bool multi)))
+    (('workers workers) (conf-set! '(server workers) (->workers workers)))
     (('engine engine) (conf-set! '(server engine) (->symbol engine)))
     (('websocket websocket) (conf-set! '(server websocket) (->bool websocket)))
     (('pub pub) (conf-set! '(server pub) (basename pub)))
@@ -499,6 +545,7 @@ debug.monitor = <PATHs>")
     (('family family) (conf-set! '(host family) (->symbol family)))
     (('addr addr) (conf-set! '(host addr) addr))
     (('port port) (conf-set! '(host port) (->integer port)))
+    (('detectpath path) (conf-set! '(host detectpath) (->string/boolean path)))
     (else (error parse-namespace-host "Config: Invalid item" item))))
 
 (define (parse-namespace-backend backend)
@@ -517,7 +564,9 @@ debug.monitor = <PATHs>")
 (define (parse-namespace-session item)
   (match item
     (('path path) (conf-set! '(session path) path))
+    (('hijackcheck check) (conf-set! '(session hijackcheck) (->bool check)))
     (('backend backend) (conf-set! '(session backend) (parse-namespace-backend backend)))
+    (('i18n i18n) (conf-set! '(session i18n) (->symbol i18n)))
     (else (error parse-namespace-session "Config: Invalid item" item))))
 
 (define (parse-namespace-upload item)
@@ -545,7 +594,7 @@ debug.monitor = <PATHs>")
 
 (define (parse-namespace-cookie item)
   (match item
-     (('expire expire) (conf-set! '(cookie expires) (->integer expire)))
+    (('expires expires) (conf-set! '(cookie expires) (->integer expires)))
     (else (error parse-namespace-cookie "Config: Invalid item" item))))
 
 (define (parse-config-item item)
@@ -565,28 +614,28 @@ debug.monitor = <PATHs>")
 
 (define (parse-line line)
   (call-with-input-string
-      line
-    (lambda (port)
-      (let lp((next (read-char port)) (key? #t) (word '()) (ret '()))
-        (cond
-         ((or (eof-object? next)
-              (char=? next #\#)) ; skip comment
-          (reverse (cons (list->string (reverse word)) ret)))
-         ((char-set-contains? char-set:whitespace next)
-          ;; skip all whitespaces
-          (lp (read-char port) key? word ret))
-         ((and key? (char=? next #\.))
-          ;; a namespace end
-          (lp (read-char port) key? '() (cons (list->symbol (reverse word)) ret)))
-         ((and key? (char=? next #\=))
-          ;; value start
-          (lp (read-char port) #f '() (cons (list->symbol (reverse word)) ret)))
-         ((not key?)
-          ;; store chars of value
-          (lp (read-char port) key? (cons next word) ret))
-         (else
-          ;; store chars of key
-          (lp (read-char port) key? (cons next word) ret)))))))
+   line
+   (lambda (port)
+     (let lp((next (read-char port)) (key? #t) (word '()) (ret '()))
+       (cond
+        ((or (eof-object? next)
+             (char=? next #\#)) ; skip comment
+         (reverse (cons (list->string (reverse word)) ret)))
+        ((char-set-contains? char-set:whitespace next)
+         ;; skip all whitespaces
+         (lp (read-char port) key? word ret))
+        ((and key? (char=? next #\.))
+         ;; a namespace end
+         (lp (read-char port) key? '() (cons (list->symbol (reverse word)) ret)))
+        ((and key? (char=? next #\=))
+         ;; value start
+         (lp (read-char port) #f '() (cons (list->symbol (reverse word)) ret)))
+        ((not key?)
+         ;; store chars of value
+         (lp (read-char port) key? (cons next word) ret))
+        (else
+         ;; store chars of key
+         (lp (read-char port) key? (cons next word) ret)))))))
 
 (define (init-inner-database-item)
   (define dbd (get-conf '(db dbd)))
@@ -641,12 +690,29 @@ debug.monitor = <PATHs>")
 ;; And init-server should be called automatically.
 (define current-conf-file (make-parameter #f))
 (define* (current-myhost #:key (for-header? #f))
-  (let* ((host (get-conf '(host name)))
+  (define (domain file)
+    (if (and file (file-exists? file))
+        (string-trim-both (call-with-input-file file read-line))
+        #f))
+  (define (get-domain-name)
+    (cond
+     ((domain (get-conf '(host detectpath))) => identity)
+     ((getenv "DOMAIN_NAME") => identity)
+     ((file-exists? "/proc/sys/kernel/domainname")
+      (let ((domain (string-trim-both
+                     (call-with-input-file "/proc/sys/kernel/domainname" read-line))))
+        (if (string=? "(none)" domain)
+            #f
+            domain)))
+     (else #f)))
+  (let* ((host (or (get-domain-name) (get-conf '(host name))))
          (real-host (if host host (get-conf '(host addr))))
          (port (get-conf '(host port))))
     (if for-header?
         (cons real-host #f)
-        (format #f "http://~a:~a" real-host port))))
+        (if host
+            (format #f "http://~a" real-host)
+            (format #f "http://~a:~a" real-host port)))))
 
 (define (init-config)
   (define conf-file (current-conf-file)) ; user specified config
